/*	UEFTapePCMInputStream.java 1.20 Release Candidate
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version 2
 *	of the License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *	(C) 2004 Richard Brown
 *
 */
 
package com.rwbrown.audio.spi.uef;

import java.io.InputStream;
import java.io.PushbackInputStream;
import java.io.IOException;

import javax.sound.sampled.AudioFormat;

class UEFTapePCMInputStream extends InputStream {
	
	/*
	 *	Our AudioFormat object, representing a 44.1Khz mono
	 *	8-bit big-endian unsigned stream. 
	 *	NB. This has to be created before we try to
	 *	initialise the default wave profile.
	 */
	 
	protected static final AudioFormat AUDIOFORMAT	=	
		new AudioFormat(44100, 8, 1, false, true);
		
	 
	/*
	 *	This is the template we use to provide ourselves with the
	 *	capability of handling high tone with dummy byte blocks.
	 *	The current stratergy is to simulate two high tone tape 
	 *	blocks and one implicit tape block.
	 *	To do this we overwrite elements 
	 *	0x06, 0x07, 0x015 and 0x16.
	 *	
	 */
	 	
	private static final byte[] HTN_DUMMY	= new byte[] {	
		0x10, 0x01, 0x02, 0x00, 0x00, 0x00,
		0x00, 0x00, 
		0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 
		(byte)0xAA,
		0x10, 0x01, 0x02, 0x00, 0x00, 0x00,
		0x00, 0x00
	};
	
	/*
	 *	This is a template we use to simulate defined
	 *	tape data blocks with implicit tape data blocks.
	 *
	 */
	
	private static final byte[] IMP_DUMMY	= new byte[] {
		8, (byte)'N', 1
	};
	
	/*
	 *	TODO. COMMENT.
	 *
	 */
	 
	private static final int[] VOLUME		= new int[] {
		120, 90, 0
	};
	
	/*
	 *	Our PushbackInputStream object. We primarily
	 *	use this to simulate the more difficult
	 *	tape blocks.
	 *
	 */
	 
	protected PushbackInputStream in;	
	
	/*
	 *	version - The version of UEFFile being read.
	 */
	
	private int version;
	
	/*
	 *	Basic information about the current tape chunk.
	 *	chunkType - The current chunk type being read.
	 *	chunkSize - The size of the current chunk.
	 *
	 */
	 
	private int chunkType;
	private int chunkSize;
	
	/*
	 *	TODO. COMMENT.
	 *
	 */
	
	private int byteValue;
	private int bitValue;
	private int parityValue;
	
	/*
	 *	These values change from chunk to chunk.
	 *
	 *	TODO. FINISH COMMENT.
	 *
	 */
	
	private int bitCount;
	private int bit;
	
	private int start;
	private int data;
	private int parity;
	private int stop;
	private int bitsPerByte;
	
	private char parityType;
	private boolean stopWave;
	
	/*
	 *	Information about the security mode.
	 *	security - Set to true if we are processing
	 *	security waves.
	 *	securityStart - Set to 'P' if the first wave
	 *	only uses the last Pi radians, otherwise set to
	 *	'W'.
	 *	securityEnd - Set to 'P' if the last wave
	 *	only uses the first Pi radians, otherwise set to
	 *	'W'.
	 *
	 */
	 	 
	private boolean security;
	private char securityStart;
	private char securityEnd;
	
	
	/*
	 *	TODO. COMMENT.
	 *
	 */
	
	private int frame;
	private int endFrame;
	
	
	/*
	 *	These values remain constant throughout
	 *	unless changed by a chunk.
	 *	baudRate - Our baud rate, typically 1200.
	 *	In the case of UEFs the baud rate is
	 *	'the number of bits per second'.
	 *	baudFrames - The number of frames required
	 *	to make an interval of length 1/baudRate
	 *	seconds.
	 *	phase - The phase of our wave generator,
	 *	this is usually 180 degrees at the outset.
	 *	
	 */
	
	private double baudRate;
	private int baudFrames;
	private int phase;
	
	
	/*
	 *	This method performs all of the tape chunk
	 *	logic. It's assumed that this method is
	 *	called from a synchronized method.
	 *
	 */
	
	private boolean hasNext() throws IOException {
		
		bit			= 0;
		bitCount	= 0;
		
		start		= 0;
		data		= 8;
		parity		= 0;
		stop		= 0;
		bitsPerByte	= 0;
		
		parityType	= 'N';
		stopWave	= false;
		
		security		= false;
		securityStart	= 'W';
		securityEnd		= 'W';
		
		while (true) {
			
			chunkType	= in.read() | in.read()<<8;
			chunkSize	= in.read() | in.read()<<8 |
				in.read()<<16 | in.read()<<24;
				
			if ((chunkType|chunkSize)<0)
				return false;
			
			switch (chunkType) {
				
				case 0x0100:
				/*
				 *	Implicit start/stop bit tape block.
				 *	We implement this now by providing
				 *	the code with a simulated &104 block
				 *	header.
				 *
				 */
				in.unread(IMP_DUMMY);
				chunkSize	+= 3;
				case 0x0104:
				/*
				 *	Defined data format tape block.	
				 *	Read in number of data bits, parity type
				 *	and number of stop bits.
				 *	Specification permits a negative number of 
				 *	stop bits to help with Acorn Atom tape images.
				 *	When the number of stop bits is negative then
				 *	we are required to output a number of waves
				 *	as stop bits. We deal with this case by
				 *	converting this number back into the number
				 *	of bits required. If the the number of stop bits
				 *	is an even negative number then we simply have
				 *	half that number of stop bits. If the number is odd
				 *	then we have half that number plus one stop bits
				 *	and we only need half the last stop bit. 
				 *
				 */
				start		= 1;
				data		= in.read();
				parityType	= (char)in.read();
				stop		= (byte)in.read();
				if (stop<0) {
					stopWave	= (--stop%2==0);
					stop		/= -2;
				}
				if (parityType =='O' || parityType =='E')
					parity	= 1;
				bitsPerByte	= start+data+parity+stop;
				bitCount	= bitsPerByte*(chunkSize-3);
				return true;
				
				case 0x0102:
				/*
				 *	Explicit tape block.
				 *	We do not output any start/stop bits
				 *	ourselves.
				 *
				 */
				bitsPerByte		= data;
				bitCount		= bitsPerByte*chunkSize;
				if (version>=9)	
					bitCount	-= in.read();
				return true;
				
				case 0x0112:
				/*
				 *	Baudwise gap tape block.
				 *	NB. Prior to version 9 the interval
				 *	used to be stored in milliseconds. Since 
				 *	version 9 we use new units. See the
				 *	UEF Specification for details.
				 *	In order to indicate silence we
				 *	make bitsPerByte negative.
				 *
				 */
				bitsPerByte		= -1;
				case 0x0110:
				/*
				 *	High tone tape block.
				 *	NB. Prior to version 9 the interval
				 *	used to be stored in milliseconds. Since 
				 *	version 9 we use new units. See the
				 *	UEF Specification for details.
				 *	In order to indicate high tone we leave
				 *	bitsPerByte set to zero.
				 *
				 */
				bitCount		= in.read() | 
					in.read()<<8;
				if (version<9)
					bitCount	*= 
						2*baudRate/1000;
				return true;
				
				case 0x0111:
				/*
				 *	High tone with dummy byte tape block.
				 *	Implement this as two high tone chunks
				 *	and a single implicit data chunk by
				 *	using the unread methods available to us.
				 *
				 */
				byte[] buffer	= (byte[])HTN_DUMMY.clone();
				buffer[0x00]	= (byte)in.read();
				buffer[0x06]	= (byte)in.read();
				buffer[0x15]	= (byte)in.read();
				buffer[0x16]	= (byte)in.read();
				in.unread(buffer);
				break;
				
				case 0x0113:
				/*
				 *	Change of baud rate tape block. 
				 *	As we do not actually find any audio data
				 *	to output we do not return from this method.
				 *	MakeUEF 1.1 can produce files that specifies
				 *	that the new baud rate should be zero. If
				 *	this happens we simply ignore the block.
				 *
				 */
				float newBaud	= Float.intBitsToFloat(
					in.read() | in.read()<<8 | 
					in.read()<<16 | in.read()<<24);
				if (newBaud==0.0f) break;
				baudRate		= newBaud;
				baudFrames		= (int)(
					AUDIOFORMAT.getSampleRate()
					/baudRate
				);
				break;
				
				case 0x0114:
				/*
				 *	Security waves.
				 *
				 */
				bitsPerByte		= data;
				bitCount		= in.read() | 
					in.read()<<8 | in.read()<<16;
				securityStart	= (char)in.read();
				securityEnd		= (char)in.read();
				security		= true;
				return true;
				
				case 0x0115:
				/*
				 *	Phase change block.
				 *	The definition of this chunk changed
				 *	between drafts 20 and 21 of the 0.10
				 *	standard. Prior to draft 20, this used 
				 *	to be a polarity inversion, that is to 
				 *	say a phase change of 180 degrees and 
				 *	the chunk had no body.
				 *
				 */
				if (chunkSize==0)
					phase	+= 180;
				else
					phase	= in.read() | 
						in.read()<<8;
				phase		%= 360;
				break;
				
				case 0x0116:
				/*
				 *	Floating point gap.
				 *	Specifies the gap in seconds.
				 *	In order to indicate silence we
				 *	make bitsPerByte negative.
				 *
				 */
				bitsPerByte		= -1;
				float gap		= Float.intBitsToFloat(
					in.read() | in.read()<<8 | 
					in.read()<<16 | in.read()<<24);
				bitCount		= (int)(2*baudRate*gap);
				System.out.println(bitCount);
				return true;
				
				default:
				in.skip(chunkSize);
				
			}
		}
	}
	
	/*
	 *	Returns -1 if no more tape data can be found.
	 *	TODO. Complete comment. 2 for silence;
	 *
	 */
	 
	private int nextBit() throws IOException {
		
		if (bit==bitCount)
			do {
				if (!hasNext())
					return -1;
			} while (bitCount==0);	
		
		int result	= 0x01;
		
		endFrame	= baudFrames/2;
		frame		= 0x00;
		
		if (bitsPerByte<0) {
		
			result			= 0x02;
		
		} else if (bitsPerByte>0) {
			
			int bitIndex	= bit % bitsPerByte;
			endFrame		= baudFrames;
			
			if (bitIndex==0) {
				byteValue	= in.read();
				parityValue	= 0;	
			} 
			
			if (bitIndex-start<0)
				
				result			= 0x00;
			
			else if (bitIndex-start<data) {
				
				result 			&= 
					byteValue>>(bitIndex-start);
				parityValue		^= result;
			
			} else if (bitIndex-start<data+parity) {
				
				if (parityType=='E')
					result		= parityValue;
				else if (parityType=='O')
					result		= parityValue^0x01;
			
			} else {
				
				result			= 0x01;
				if (bitIndex==bitsPerByte-1 && stopWave)
					endFrame	/= 2; 
			}
			
			if (security) {
				if ((result^=0x01)==0x01)
					endFrame	/= 2;
				if (bit==0 && securityStart=='P') {
					frame		= endFrame/2;
				} else if (bit==bitCount-1
					&& securityEnd=='P') {
					endFrame	/= 2;
				}
			}
		}
		
		bit++;
		
		return result;
	}
	
	public void close() throws IOException {
		super.close();
		in.close();
	}
		
	public synchronized int read() throws IOException {
		if (frame==endFrame && (bitValue=nextBit())<0)
			return -1;
		return 0x80-(int)(VOLUME[bitValue]*Math.sin(
			2*Math.PI*(1+bitValue)*frame++/baudFrames
			+Math.toRadians(phase)
			)
		);
	}
	
	public synchronized int read(byte[] b, int off, int len) 
		throws IOException {
		return super.read(b, off, len);
	}
	
	/*
	 *
	 *	Our constructor.
	 *
	 */
	 
	UEFTapePCMInputStream(InputStream stream) throws IOException {
		super();
		in				= new PushbackInputStream(
			stream, 1024);
		byte buffer[]	= new byte[10];
		for (int i=0; i<10;)
			buffer[i++]	= (byte)in.read();
		String header	= new String(buffer, "us-ascii");
		if (!header.trim().equals("UEF File!")) throw
			new IOException ("Not a UEF File!");
		version			= in.read() | in.read()<<8;
		baudRate		= 1200.f;
		baudFrames		= (int)(
			AUDIOFORMAT.getSampleRate()
			/baudRate
		);	
	}
}